#pragma once

#include <fmt/core.h>
#include <SFML/Graphics.hpp>
#include <SFML/Graphics/Image.hpp>
#include <numbers>
#include <algorithm>
#include <cmath>
#include "matrix.hpp"
#include <cstdlib>
#include <array>
#include "dbc.hpp"
#include <memory>

using matrix::Matrix;

struct Sprite {
  double x;
  double y;
  int texture;
  // ZED: this should be a separate transform parameter
  double elevation=0;
  int uDiv=1;
  int vDiv=1;
};

using RGBA = uint32_t;

struct TexturePack {
  int NUM_SPRITES=1;
  int NUM_TEXTURES=11;
  int TEXTURE_WIDTH=256; // must be power of two
  int TEXTURE_HEIGHT=256; // must be power of two

  std::vector<std::vector<uint32_t>> images;
  std::vector<Sprite> SPRITE{{4.0, 3.55, 8}};
  const int floor = 3;
  const int ceiling = 6;

  void load_textures();
  std::vector<uint32_t> load_image(const char *filename);
  Sprite &get_sprite(size_t sprite_num);
  std::vector<uint32_t>& get(size_t num);
};

struct Raycaster {
  int PITCH=0;

  TexturePack textures;
  double posX = 0;
  double posY = 0;

  // initial direction vector
  double dirX = -1;
  double dirY = 0;

  // the 2d raycaster version of camera plane
  double planeX = 0;
  double planeY = 0.66;
  sf::Texture view_texture;
  sf::Sprite view_sprite;

  //ZED: USE smart pointer for this
  std::unique_ptr<RGBA[]> pixels = nullptr;

  int $width;
  int $height;
  sf::RenderWindow& $window;
  Matrix& $map;
  std::vector<int> spriteOrder;
  std::vector<double> spriteDistance;
  std::vector<double> ZBuffer; // width

  Raycaster(sf::RenderWindow& window, Matrix &map, int width, int height);

  void draw_pixel_buffer();
  void clear();
  void cast_rays();
  void draw_ceiling_floor();
  void sprite_casting();
  void sort_sprites(std::vector<int>& order, std::vector<double>& dist, int amount);
  void render();

  bool empty_space(int new_x, int new_y);

  void run(double speed, int dir);
  void rotate(double speed, int dir);
  void position_camera(float player_x, float player_y);

  void set_position(int x, int y);
  inline size_t pixcoord(int x, int y) {
    return ((y) * $width) + (x);
  }

};
